#!/usr/bin/env python
"""

"""
# Author: Harco Kuppens


import sys, re, pprint  # modules from standard lib (python 2.6 and later)
import yaml
import argparse
import tempfile
import shutil


import uppaal2bcg
import uppaal2aut
import model_parser 
import register 

# load libraries from ./lib
import lib.util,lib.check,lib.run,lib.error


# TODO: add other lts types with nice lts2lts script
import learnlib_dot2jtorx_aut,jtorx_aut2cadp_aut
#import lts2lts


# explicit define public api for when used as module
__all__ = [ 'compare' ]
# note: main and argparsing are just helpers for the function give_user_acces_to_dir_using_acl



def args_parsing(args):
    import os.path
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('--keep-tmp-files', action='store_true', help='keep temporary files generated for doing comparison')
    argparser.add_argument('modelfile1', help='model file which can be one of : uppaal model,cadp aut, jtorx aut, dot')
    argparser.add_argument('modelfile2', help='model file which can be one of : uppaal model,cadp aut, jtorx aut, dot')
    argparser.add_argument('range_min', help='minimum value of range used for input params when flattening the model')
    argparser.add_argument('range_max', help='maximum value of range used for input params when flattening the model')

    parse_result=argparser.parse_args(args[1:])

    keeptmp=parse_result.keep_tmp_files

    # handle input/output file argument
    modelfile1=os.path.abspath(parse_result.modelfile1).replace('\\','/')
    modelfile2=os.path.abspath(parse_result.modelfile2).replace('\\','/')
    # Note: on windows also interprets / as dir sep, and returns path with windows \


    lib.check.check_isreadablefile(modelfile1)
    lib.check.check_isreadablefile(modelfile2)

    range_min=parse_result.range_min
    range_max=parse_result.range_max

    return  [modelfile1,modelfile2,range_min,range_max,keeptmp]


def model2bcg(input,outdir,name,range_min,range_max,keeptmp):

    output_str=""

    bcgfile=outdir + '/' + name + '.bcg'
    if input.endswith('xml'):
       uppfile=""
       modeltype=model_parser.get_modeltype(input)
       if modeltype == "uppaal":
           uppfile=input
       elif modeltype == "register":
           uppfile=outdir + '/' + name + '.uppaal.xml'
           regmodel=register.from_xml(input)
           regmodel.to_uppaalfile(uppfile)
       else:
           print "error: this xml input model file is not supported. Xml model file must be either an 'uppaal' or 'register' model."
           exit(1)

       output_str+=uppaal2bcg.uppaal2bcg(uppfile,bcgfile,range_min,range_max,keeptmp)

    if input.endswith('bcg'):
       shutil.copy(input,bcgfile)

# TODO: add other lts types with nice lts2lts script
    if input.endswith('cadp.aut'):
      lib.run.run_command([ 'bcg_io',  input,bcgfile ],cwd=outdir)
    elif input.endswith('jtorx.aut'):
      cadp_aut_file=outdir + '/' + name + '.cadp.aut'
      jtorx_aut2cadp_aut.jtorx_aut2cadp_aut(input,cadp_aut_file)
      lib.run.run_command([ 'bcg_io',  cadp_aut_file,bcgfile ],cwd=outdir)
    elif input.endswith('.aut'):     # if not specified assumes tomte :  I/O and '_' separator
      cadp_aut_file=outdir + '/' + name + '.aut'
      uppaal2aut.tomte_aut2cadp_aut(input,cadp_aut_file)
      lib.run.run_command([ 'bcg_io',  cadp_aut_file,bcgfile ],cwd=outdir)
    elif input.endswith('dot'):
      jtorx_aut_file=outdir + '/' + name + '.jtorx.aut'
      learnlib_dot2jtorx_aut.dot2jtorxaut(input,jtorx_aut_file)
      output_str+=model2bcg(jtorx_aut_file,outdir,name,range_min,range_max,keeptmp)



    return output_str

def generateModelFiles(modelfile1,modelfile2,range_min,range_max,tmpdir,keeptmp=False):
    line="\n-----------------------------------------------\n" 
    output_str="flatten first model to a cadp bcg file" 
    output_str+=line
    output_str+=model2bcg(modelfile1,tmpdir,'model1',range_min,range_max,keeptmp)
    output_str+="flatten second model to a cadp bcg file" 
    output_str+=line
    output_str+=model2bcg(modelfile2,tmpdir,'model2',range_min,range_max,keeptmp)
    return output_str.strip()


def compareModels(tmpdir,reverse=False):
    line="\n-----------------------------------------------\n" 
#    output_str="flatten first model to a cadp bcg file" 
#    output_str+=line
#    output_str+=model2bcg(modelfile1,tmpdir,'model1',range_min,range_max,keeptmp)
#    output_str+="flatten second model to a cadp bcg file" 
#    output_str+=line
#    output_str+=model2bcg(modelfile2,tmpdir,'model2',range_min,range_max,keeptmp)

    model1 = 'model1.bcg'
    model2 = 'model2.bcg'
    if reverse:
        (model1,model2)=(model2,model1)

    output_str="compare " + model1 + " with " + model2 + " with cadp caesar" 
    output_str+=line
    output_str+=lib.run.run_command_with_output(['time', 'bcg_open', model1 , 'bisimulator', '-diag', '-bfs', '-strong',  model2 ],cwd=tmpdir)
    return output_str.strip()
   

def getTraceFromText(s): 
        # get Counter example from text
        start=s.rfind('<start state>')+14
        end=s.rfind('<goal state>')
        s.rfind('<goal state>')
        ioactions=s[start:end].translate(None,'\x01\x02').strip().split('\n')  #"\x01"
        # relabel labels from cadp format to tomte format
        ioactions=[label.strip('"').replace(' !','_') for label in ioactions]
    
        #pprint.pprint(ioactions)
    
        # remove outputs by only allowing even entries
        inputs=ioactions[::2]
        outputs=ioactions[1::2]
   
        # differening input/output has extra text 'Present in <modelname>.bcg'
        # where modelname can be eithre model1 or model2 depending on the order
        # we now remove this 
        if len(inputs) > len(outputs)   :
            # differ by input 
            inputs[-1]=inputs[-1].replace('Present in model2.bcg: ','')
            inputs[-1]=inputs[-1].replace('Present in model1.bcg: ','')
        else: 
            # differ by output 
            outputs[-1]=outputs[-1].replace('Present in model2.bcg: ','')
            outputs[-1]=outputs[-1].replace('Present in model1.bcg: ','')

        #pprint.pprint(inputs)
        #pprint.pprint(outputs)
    
        return (inputs,outputs)

def compare(modelfile1,modelfile2,range_min,range_max,keeptmp=False):
    tmpdir=tempfile.mkdtemp(prefix="sut_compare_")


    #print "tmpdir: ",tmpdir
    #print  output_str

    s=generateModelFiles(modelfile1,modelfile2,range_min,range_max,tmpdir)
    output_str=s
    s=compareModels(tmpdir)
    output_str=output_str+s
    equal=s.endswith('TRUE')

    counterExample=None
    outputsModel1=None
    outputsModel2=None


    if not equal:
        (inputs,outputs)=getTraceFromText(s) 
        counterExample=inputs
        outputsModel2=outputs

        # redo compare with models switched to get outputs of model1
        s=compareModels(tmpdir,True)
        output_str=output_str+s
        (inputs,outputs)=getTraceFromText(s) 
        #counterExample=inputs
        outputsModel1=outputs

    if not keeptmp:
        shutil.rmtree(tmpdir)

    # remove the control chars from the cadp output
    output_str=''.join(c for c in output_str if ord(c) >= 10)

    return {'equal':equal,'inputsCounterExample':counterExample,'outputsModel1':outputsModel1,'outputsModel2':outputsModel2,'fullOutput':output_str}

def main(args):
    """

    """

    (modelfile1,modelfile2,range_min,range_max,keeptmp)=args_parsing(args)



    result= compare(modelfile1,modelfile2,range_min,range_max,keeptmp)
#    print result['fullOutput']
    import json
    print json.dumps(result, sort_keys=True, indent=4)
    #print json.dumps(result)


if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt as e:
        print >> sys.stderr, "\n\nExiting on user cancel."
        sys.exit(1)

